// constants.h 
#ifndef CONSTANTS_H
#define CONSTANTS_H

// Probability of a packet being selected
// for the packet cache.
#define SAMPLE_PROBABILITY 0.03
//#define SAMPLE_PROBABILITY 0.0

// The size of the packet cache
#define P_CACHE_SIZE 32768
#define P_CACHE_MAX_SAMPLE 4096

// For debugging niceties
#define HERE { fprintf(stderr, "%s: %d\n", __FILE__, __LINE__); }

// Constants for C niceness
#define boolean int
#define true 1
#define false 0

// Constants for debugging
#define ALLOW_AGEING false
#define ALLOW_DEFAULT_SPLITTING true // Deprecated
#define ALLOW_SPLITTING true
#define ALLOW_MERGING true
#define DEBUG 

#define MAX_N 16


// Rates for splitting/merging
#define OPTIMAL false // enabiling optimal may work slower (but should guarantee same png selection in each run)?!
#define MAX_RATE 2.0
#define MIN_RATE 0.25  // Does this do anything?
#define MAX_MERGE_RATE 0.0

// The percentage range around which we find png splitting.
#define RANGE 0.2

// Macros

// Casting a node to png_op (Type check required beforehand)
#define NODE2PNGS(a) (&((a)->data.pngs))
// Casting a node to queue (Type check required beforehand)
#define NODE2QUEUE(a) (&((a)->data.queue))
// # of packets that went through this node during this time period
#define LAST_TICK(a) (a->count_cache[simtime % MAX_TIME])
// Is this node mature?
#define MATURE(a) ((simtime - a->born) > MAX_TIME)
// How many packets are in the cache?
#define CACHE_SIZE(a) (min(a->data.queue.packet_cache.size, MAX_PACKETS))
// For the qcap stuff
#define OK(z) assert(QCAP_OKAY(z))

// Type enumeration for nodes 
#define PNG_OP 1
#define QUEUE  2

#define MAX_QUEUES 1000    // Maximum number of queues.

// The following two lines are deprecated
#define MAX_PNGS 64        // Maximum number of pngs per queue.
#define MAX_PACKETS 1000   // Maximum number of packets to store in a
                           // time period.
                           
                          
// ------------------------------------------------------------------
// These next two constants are fundamental ones that you may want
// to change.  These in fact should probably be variables but are
// constants due to being embedded in data structures (and I would
// have to change many things to remove them.)

// This has been 10 for every published experiment (the CSE report
// and the earlier paper).
#define TICK_INTERVAL 10    // Minutes per data file.
// This is 6 * the number of hours in the training period.
// Typically, this is 18 (3 hours).
#define MAX_TIME (18)      // Should be a set parameter, but difficult
                           // to remove as a constant because it is
                           // embedded in data structures. 




#define MAX_DEPTH 128       // Maximum depth we should search for (How
                           // is this enforced?)

#define MAX_FLOWS 32        // How many flows can we monitor?

#define MAX_FILES 52560     // One year of files


/* Interactive Mode -----------------------------------------------*/

// What should we show? 
#define SHOW_PACKETS 1
#define SHOW_FLOWS   2
#define SHOW_SUMMARY 3
#define SHOW_ALL     4

// We're storing these as bits in the monitor field in node_t 
#define MONITOR_NONE         0x00
#define MONITOR_SUMMARY      0x01
#define MONITOR_PACKETS      0x02
#define MONITOR_FLOWS        0x04
#define MONITOR_ALL          0x07


// For looking at files
#define TMP_FILE "/tmp/graph.dot"
//#define DOT_PATH "/usr/bin/dot -Gsize=12,10 -Gcenter"
#define DOT_PATH "/opt/local/bin/dot -Gsize=12,10 -Gcenter"
//#define PS2PDF_PATH "/usr/bin/ps2pdf"
#define PS2PDF_PATH "/opt/local/bin/ps2pdf"


extern const char* IHV_COPYRIGHT_STRING;

// Eventually we're going to have variable length pngs

// Support for variable length pngs
//#define VARIABLE_N 

#define ETHERNET_HEADER_LENGTH 14;

#endif // CONSTANTS_H
