#ifndef SIM_DB_H
#define SIM_DB_h

#include "sim.h"

#include <time.h>
#include <stdbool.h>
#include <pcap.h>

typedef struct db db_t;
typedef struct db_tree db_tree_t;
typedef struct db_png db_png_t;
typedef struct db_capture db_capture_t;

struct db_png
{
    int p;   /*!< the positon p of the p,n-gram */
    unsigned char *n; /*!< the n-gram data */
    unsigned long len; /*!< the length of the n-gram data */
};

/*! Callback function for iterating over nodes in a database tree
  @param type QUEUE or PNG_OP
  @param id The unique ID number for the node
  @param position 1 for root, n*2 for left child, n*2+1 for right child
  @param pngs, Array of p,n-grams (NULL for non-PNG_OP types)
  @param n_pngs Number of p,n-grams (0 for non-PNG_OP types)
  @param data User-supplied data
*/
typedef void (*db_foreach_node_func)(int type, int id, int position,
                                     db_png_t *pngs, int n_pngs, void *data);

db_t* db_new(const char *host, unsigned int port, const char *dbname,
             const char *user, const char *pass);
void db_ref(db_t *db);
void db_unref(db_t *db);

bool db_connect(db_t *db);

db_tree_t* db_tree_new(db_t *db, const char *name);
void       db_tree_ref(db_tree_t *t);
void       db_tree_unref(db_tree_t *t);

bool db_tree_get_current_counters(db_tree_t *t, unsigned int *queue,
                                  unsigned int *png);

/*! Iterate over the nodes in the tree's latest revision.  The nodes are
  iterated in an order such that no node will be seen before all of its
  parents are seen.
  @param data User-specified data that will be passed to the callback function
  @return FALSE if there was no revision with nodes to read from, TRUE
          otherwise
*/
bool db_tree_foreach_node(db_tree_t *t, db_foreach_node_func func,
                          void *data);

/*! Create a new revision for the tree in the database */
void db_create_new_tick(db_tree_t *t, const time_t time,
                        unsigned int queue_counter, unsigned int png_counter);

void db_tree_lookup_capture_requests(db_tree_t *t, pcap_t *pcap);
int db_tree_num_capture_requests(db_tree_t *t);
db_capture_t* db_tree_capture_request(db_tree_t *t, unsigned int i);

void db_capture_packet(db_capture_t *cap, const struct pcap_pkthdr *h,
                       const u_char *sp);

int db_capture_node_id(db_capture_t *cap);

#endif
