#ifndef PROTOCOL_H
#define PROTOCOL_H


#include"constants.h"
#include<pcap.h>

enum ethertype_value {
    IPv4        = 0x0800,
    ARP         = 0x0806,
    RARP        = 0x8035,
    ETHERTALK   = 0x809b,
    AARP        = 0x80f3,
    IEEE8021Q   = 0x8100,
    IPX         = 0x8137,
    NOVELL      = 0x8138,
    IPv6        = 0x86dd,
    MPLSuni     = 0x8847,
    MPLSmulti   = 0x8848,
    PPPoEdisc   = 0x8863,
    PPPoEss     = 0x8864,
    ATAoE       = 0x88a2,
    LOOP        = 0x9000,
    ETHER_EXCEPTION_1 = 0x1400
};

enum ethertype_enum {
    ETHER_UNKNOWN_E,
    ETHER_DUMMY_1_E,
    ETHER_E,
    ETHER_DUMMY_2_E,
    MPLSuni_E,
    MPLSmulti_E,
    ETHER_DUMMY_3_E,
    IPX_E,
    NOVELL_E,
    ETHER_DUMMY_4_E, 
    ETHER_DUMMY_5_E,
    IPv4_E,
    ETHER_DUMMY_6_E,
    ETHERTALK_E,
    ETHER_DUMMY_7_E,
    IPv6_E,
    ETHER_EXCEPTION_1_E,
    PPPoEdisc_E,
    PPPoEss_E,
    ETHER_DUMMY_8_E,
    ETHER_DUMMY_9_E,
    ETHER_DUMMY_10_E,
    AARP_E,
    RARP_E,
    ARP_E,
    ETHER_DUMMY_11_E,
    IEEE8021Q_E,
    ATAoE_E,
    LOOP_E,
    ETHER_DUMMY_12_E,
    ETHERTYPE_ENUM_SIZE
};

enum iptype_value {
    ICMP_V    = 1,
    IGMP_V    = 2,
    TCP_V     = 6,
    UDP_V     = 17,
    EIGRP_V   = 88
};

enum iptype_enum {
   IP_UNKNOWN_E,
   IP_DUMMY_1_E,
   UDP_E,
   IP_DUMMY_2_E,
   IP_DUMMY_3_E,
   IP_DUMMY_4_E,
   IP_DUMMY_5_E,
   IP_DUMMY_6_E,
   IP_DUMMY_7_E,
   IGMP_E,
   IP_DUMMY_8_E,
   IP_DUMMY_9_E,
   EIGRP_E,
   IP_DUMMY_10_E,
   IP_DUMMY_11_E,
   IP_DUMMY_12_E,
   IP_DUMMY_13_E,
   IP_DUMMY_14_E,
   IP_DUMMY_15_E,
   IP_DUMMY_16_E,
   IP_DUMMY_17_E,
   IP_DUMMY_18_E,
   IP_DUMMY_19_E,
   IP_DUMMY_20_E,
   IP_DUMMY_21_E,
   ICMP_E,
   TCP_E,
   IP_DUMMY_22_E,
   IP_DUMMY_23_E,
   IP_DUMMY_24_E,
   IPTYPE_SIZE
};

enum udptype {
    UDP_UNKNOWN_E,
    UDP_DNS_E,
    UDP_DUMMY_1_E,
    CUPS_E,
    IPSEC_E,
    WHO_E,
    XDMCP_E,
    BITTORRENT_E,
    RTP_E,
    MSFT_SQL_E,
    NETBIOS_DGM_E,
    MSFT_MS_E,
    MDNS_E,
    UDP_DUMMY_2_E,
    GANGLIA_E,
    UDP_DUMMY_3_E,
    UDP_DUMMY_4_E,
    UDP_DUMMY_5_E,
    UDP_DUMMY_6_E,
    NBNS_E,
    RIPv1_E,
    HSRP_E,
    DHCP_E,
    UDP_DUMMY_7_E,
    UDP_DUMMY_8_E,
    SNMP_E,
    UDP_DUMMY_9_E,
    NTP_E,
    SRVLOC_E,
    UDP_DUMMY_10_E,
    UDPTYPE_SIZE
};

enum tcptype {
    TCP_UNKNOWN_E,
    TCP_DNS_E,
    MSFT_WBT_E,
    IPP_E,
    POPS_E,
    IMAPS_E,
    HTTPS_E,
    SSH_E,
    MS_STREAMING_E,
    MYSQL_E,
    SMB_E,
    DCE_RPC_E,
    MSNMS_E,
    XMPP_E,
    TCP_SOPHOS_E,
    URD_E,
    TCP_SPECIAL_LEN_E,
    RTSP_E,
    TCP_DUMMY_1_E, // SUNRPC_E,
    NBSS_E,
    IRC_E,
    NNTP_E,
    TELNET_E,
    TCP_DUMMY_2_E,
    SMTP_E,
    IMAP_E,
    CVS_E,
    POP_E,
    HTTP_E,
    FTP_E,
    TCPTYPE_SIZE
};

typedef struct {
    int total_packets;
    int ether_stats[ETHERTYPE_ENUM_SIZE];
    int ip_stats[IPTYPE_SIZE];
    int udp_stats[UDPTYPE_SIZE];
    int tcp_stats[TCPTYPE_SIZE];
} p_stats;

typedef struct {
    char* ether_names[ETHERTYPE_ENUM_SIZE];
    char* ip_names[IPTYPE_SIZE];
    char* udp_names[UDPTYPE_SIZE];
    char* tcp_names[TCPTYPE_SIZE];
} p_stats_names;

boolean isIPv4(const u_char *packet, unsigned int len);

void p_stats_init(p_stats* stats);
void p_stats_print(FILE* flog, p_stats* stats);
void p_stats_total(p_stats* total, p_stats* stats, int count);
void p_stats_print_summary(FILE* flog, p_stats* stats);
int p_stats_count_protocols(p_stats* stats);

enum ethertype_value get_ethertype(const u_char *packet, unsigned int len);
boolean classify(p_stats* stats, const u_char* packet, unsigned int len);
char* p_stats_print_pie_chart(p_stats* stats, int radius);

p_stats_names* p_stats_get_db_names();

#endif // PROTOCOL_H
