#include"sim.h"
#include"stats.h"
#include"db.h"
#include<glib.h>
#include<float.h>
#include<string.h>
#include<math.h>
#include<netinet/in.h>
#include<arpa/inet.h>
#include<pcap.h>
#include<stdarg.h>
#include"flow.h"
#include<time.h>
#include"protocol.h"
#include"graphics.h"
#include"vector.h"
#include"cache.h"
#include<mysql/mysql.h>

// Temps

// Globals  ----------------------------------------------------------
extern char* optarg;
extern int   optind, opterr, optopt;
extern FILE* yyin;

int      N      = 2;    // This should be the default N (perhaps we should calculate this somehow?)

FILE*    flog   = NULL;
FILE*    fgraph = NULL;

char*    queue_file = NULL;
static char**   filenames;

const config_t *config = NULL;
int     datafile = 0;

int     initial_queues = 50;
int     initial_pngs = 1000;
int     queue_count = 0;
int     packet_counts[MAX_TIME];
int     total_packets = 0;
int     max_simtime_packets = 0;
node_t* head = NULL;

int     ip_offset = 0;

pcap_t* pcap = NULL;
bool    end_loop = false;

static node_t*  queues[MAX_QUEUES];

int num_queues = 0;

static int png_counter;
static int queue_counter;

int      simtime = 0;
int      tick_packet_count = 0;

long int     last_tick = 0;
int     last_capture_check = 0;

double      prob = 0.2;     // Sample probability

packet_cache_t* cache = NULL;

boolean     old_algorithm = false;
boolean     globals = true;

void print_queue_array(FILE* file);
void print_graph(FILE* file);
int sum_max_simtime(node_t* q);
int png_match(png_t* png, const u_char* packet, unsigned int len);
void sanity();
void node_queue_free(node_t* node);

p_stats global_stats;
p_stats tick_stats[MAX_TIME];

db_t *db = NULL;
db_tree_t *db_tree = NULL;

jim_vector_t* cluster_visitors;

void head_set(node_t* node)
{
    if(config->verbose) fprintf(flog, "Setting head to n%d.\n", node->id);
    head = node;
}

inline float percent(int x, int y);
inline void queues_inc() { num_queues++; }
inline void queues_dec() { num_queues--; }

/*
void init_packet_cache(queue_t* queue, int free)
{
    int i;
    for(i=0; i < MAX_PACKETS; i++) {
        if( free && queue->packet_cache.packets[i] != NULL )
            qcap_packet_free(queue->packet_cache.packets[i]);
        queue->packet_cache.packets[i] = NULL;
    }
    queue->packet_cache.size = 0;

}
*/


void count_mono_nodes(node_t* node, void* data)
{
    assert(node->type == QUEUE);
    int* count = (int*)data;
/*
    p_stats totals; p_stats_init(&totals);
    p_stats_total(&totals, node->data.queue.stats, MAX_TIME); 
*/
    if(p_stats_test_mono(&(node->data.queue.stats[simtime % MAX_TIME]))) (*count)++;
 //   if(p_stats_test_mono(&totals)) (*count)++;
}

void total_mono_packets(node_t* node, void* data)
{
    assert(node->type == QUEUE);
    int* count = (int*)data;
/*
    p_stats totals; p_stats_init(&totals);
    p_stats_total(&totals, node->data.queue.stats, MAX_TIME);
*/
  //  if(p_stats_test_mono(&totals))
    if(p_stats_test_mono(&(node->data.queue.stats[simtime % MAX_TIME])))
       (*count) += node->count_cache[simtime % MAX_TIME]; 
}

void node_queue_init(node_t* node, int free)
{
    int i;
    assert(node != NULL && node->type == QUEUE);

    node->frozen  = false;
    node->left    = NULL;
    node->right   = NULL;
    node->born    = simtime;
    node->monitor = MONITOR_NONE;
    //init_packet_cache(NODE2QUEUE(node), free);
 
    // For the new packet cache stuff
    if(cache) packet_cache_free(cache, NODE2QUEUE(node));

    for(i=0; i < MAX_TIME; i++)
        node->count_cache[i] = 0;

    node->id = id_queue_get_next();

    p_stats_init(&(node->data.queue.stats_total));
    for(i=0; i < MAX_TIME; i++)
        p_stats_init(&(node->data.queue.stats[i]));
}


node_t* node_queue_create(node_t* parent)
{
    int i;
    node_t* ret = (node_t*)malloc(sizeof(node_t));
    ret->type   = QUEUE;

    assert(parent == NULL || (parent != NULL && parent->type == PNG_OP));

    node_queue_init(ret, false);
    ret->parent = parent;

    assert(num_queues < MAX_QUEUES);
    queues[num_queues] = ret;
    queues_inc();
    return ret;
}


// Stack -------------------------------------------------------------
sig_stack_t* stack_create()
{
    int i;
    sig_stack_t* ret = (sig_stack_t*)malloc(sizeof(sig_stack_t));
//    for(i=0; i < MAX_DEPTH; i++) ret->stack[i] = NULL;
    ret->size = 0;
    return ret;
}

void stack_free(sig_stack_t* stack)
{
    free(stack);
}

void stack_push(sig_stack_t* stack, boolean match, png_op_t* pngs)
{
    assert(stack->size < MAX_DEPTH);
    stack->stack[stack->size].pngs = pngs;
    stack->stack[stack->size].match = match;
    stack->size++;
}

sig_stack_elem_t stack_pop(sig_stack_t* stack)
{
    sig_stack_elem_t ret = stack->stack[stack->size-1];
    stack->size--;
    return ret;
}

void print_queue_summary(FILE* file)
{
}

void print_queue(FILE* file, node_t* node, sig_stack_t* stack)
{
    int i;
    assert(node != NULL && node->type == QUEUE && stack->size >= 0);

    p_stats totals; p_stats_init(&totals);
    p_stats_total(&totals, node->data.queue.stats, MAX_TIME);

    char* state = "N";

    if(MATURE(node))
    {
        if(queue_mergable(node))
            state = "M";
        else if (queue_splittable(node))
            state = "S";
    }
    else
        state = "I";

    {
        if(node == node_default())
        {
            fprintf(file, "n%d (default): %s %d %d ( %.2f ) %d ( %.2f ) %d  ",
                       node->id, state, node->born,
                       LAST_TICK(node), percent(LAST_TICK(node), tick_packet_count), 
                       sum_max_simtime(node), percent(sum_max_simtime(node), max_simtime_packets),
                       p_stats_count_protocols(&totals));
        }
        else
        {
            fprintf(file, "n%d: %s %d %d ( %.2f ) %d ( %.2f ) %d  ",
                       node->id, state, node->born,
                       LAST_TICK(node), percent(LAST_TICK(node), tick_packet_count), 
                       sum_max_simtime(node), percent(sum_max_simtime(node), max_simtime_packets), 
                       p_stats_count_protocols(&totals));
        }

        // Now print the signature
        if(stack->size > 0)
        {
            fprintf(file, "\"");
            for(i=0; i < (stack->size)-1; i++)
            {
                if(!(stack->stack[i].match))
                    fprintf(file, "!");
                fprintf(file, "(");
                pngs_print(file, stack->stack[i].pngs);
                fprintf(file, ") && ");
            }
 
            if(!(stack->stack[i].match))
                fprintf(file, "!");
            fprintf(file, "(");
            pngs_print(file, stack->stack[(stack->size)-1].pngs);
            fprintf(file, ")");
                fprintf(file, "\"");
        }
        fprintf(file, "\n");
    }      
    if(cache)
        packet_cache_stats(file, cache, NODE2QUEUE(node));
}

void print_queues(FILE* file, node_t* node, sig_stack_t* stack)
{
    assert(node != NULL);
    
    if(node->type == QUEUE)
    {
        print_queue(file, node, stack);     
    }
    else
    {
        stack_push(stack, true, &(node->data.pngs)); 
        print_queues(file, node->left, stack);
        stack_pop(stack);

        stack_push(stack, false, &(node->data.pngs)); 
        print_queues(file, node->right, stack);
        stack_pop(stack);
    }
}

void queues_print(FILE* file)
{
    sig_stack_t* stack = stack_create();
    print_queues(file, head, stack);
    stack_free(stack);
}


void add_packet(queue_t* queue, const u_char* packet, unsigned int len)
{
    int ret;
    double val;

    // For the new packet cache infrastructure
    // This will make things slow

    val = drand48();
    if(val < prob)
    {
        /* FIXME this is a bug! casting the queue to the node_t is not getting
           you the id! */
        packet_cache_insert(cache, queue, packet, len, ((node_t*)queue)->id);
    }
}

int png_op_match(png_op_t* pngs, const u_char* packet, unsigned int len)
{
    int i;
    for(i=0; i < pngs->png_size; i++)
    {
        if(png_match(pngs->pngs[i], packet, len))
            return true;
    }

    return false;
}




int match(node_t* node, const u_char* packet, unsigned int len)
{
    int node_id;

    assert(node->type == PNG_OP || node->type == QUEUE);
    if(node == NULL)
    {
        fprintf(stdout, "ERROR!  match passed NULL.\n");
        exit(1);
    }
    else if(node->type == PNG_OP)      // Match against pngs
        if(png_op_match(&(node->data.pngs), packet, len))
        {
            node_id = match(node->left, packet, len);
        }
        else
        {
            node_id = match(node->right, packet, len);
        }
    else                    // Add it to the queue
    {
        // Some sort of generalized pattern should be used for the
        // following things.  CLEANUP
        add_packet(&(node->data.queue), packet, len);
        if (node->count_cache[simtime % MAX_TIME] > 5000000)
        {
            fprintf(flog, "ERROR: node->count_cache[simtime] = %d\n", 
                    node->count_cache[simtime % MAX_TIME]);
            fprintf(flog, "This is abnormally high.\n");
            fprintf(flog, "Perhaps you should debug your code.\n");
            exit(0);
        }


        // This should probably be more general, and allow any
        // sort of match.
        if(flow_matches(packet, len))
        {
            flow_t* flow = flow_from_packet(packet, len);
            fprintf(stdout, "Packet with flow: ");
            flow_print(stdout, flow);
            fprintf(stdout, " is handled by queue N%d.\n", node->id);
        }

        // For command_monitor
        if(node->monitor & MONITOR_PACKETS != 0)
        {
            flow_t* flow = flow_from_packet(packet, len);
            fprintf(stdout, "Packet from node: N%d -> ", node->id);
            if(flow)
            {
                flow_print(stdout, flow);
                free(flow);
            }
            else
                fprintf(stdout, " <NOT TCP | UDP>.");
            fprintf(stdout, "\n");
        }

        classify(&(node->data.queue.stats_total), packet, len);
        classify(&(node->data.queue.stats[simtime % MAX_TIME]), packet, len);

        node_id = node->id;
    }
    node->count_cache[simtime % MAX_TIME] += 1;

    return node_id; /* return the matched node's id */
}

void node_iter_func(int type, int id, int position,
                    db_png_t *pngs, int n_pngs, void *data)
{
    printf("%s node %05d at position %d\n",
           (type == PNG_OP ? "p,n-gram" : "queue"), id, position);
    if (type == PNG_OP) {
        gint i;
        for (i = 0; i < n_pngs; ++i) {
            unsigned long j;
            printf("  png %d: %4d,0x", i, pngs[i].p);
            for (j = 0; j < pngs[i].len; ++j)
                printf("%02x", pngs[i].n[j]);
            printf(" (%d)\n", pngs[i].len);
        }
    }
}

void initialize_globals(const config_t *c)
{
    int i;
    node_t* node;

    config = c;

    if (config->database_output) {
        db = db_new(config->database_host,
                    config->database_port,
                    config->database_name,
                    config->database_user,
                    config->database_pass);
        if (db) {
            if (db_connect(db)) {
                db_tree = db_tree_new(db, config->database_tree);
            } else {
                db_unref(db);
                db = NULL;
                fprintf(stderr, "Unable to connect to the database.\n");
            }
        } else
            fprintf(stderr, "Error initializing MySQL\n");
    }

    if (config->dot_output)
        if ((fgraph = fopen(config->dot_file, "w")))
            p_stats_print_key(fgraph);
    
    if (!fgraph && !db) {
        fprintf(stderr, "No valid output. Use --db or --dot. Exiting.\n");
        exit(EXIT_FAILURE);
    }

    p_stats_init(&global_stats);

    for(i=0; i < MAX_TIME; i++)
        packet_counts[i] = 0;

    for(i=0; i < MAX_QUEUES; i++)
        queues[i] = NULL;

    cache = packet_cache_init(P_CACHE_SIZE);

    // For iterating over the clusters.
    cluster_visitors = jim_vector_create();

    if (!db_tree ||
        !db_tree_get_current_counters(db_tree, &queue_counter, &png_counter))
    {
        png_counter = 0;
        queue_counter = 10000;
    }

    node = node_queue_create(NULL);
    node_queue_init(node, false);
    head_set(node);
}

void free_globals()
{
    db_unref(db);
    db_tree_unref(db_tree);
    packet_cache_delete(cache);
}


void read_line(char* line, FILE* file)
{
    fgets(line, 256, file); 
}


png_t* png_create()
{
    png_t* ret;

    ret = (png_t*)malloc(sizeof(png_t));

    ret->val = (u_char*)malloc(sizeof(u_char)*N);
    ret->offset = 0;
    ret->match = 0;
    
    return ret;
}

png_t* png_init(png_t* png, int offset, u_char* vals)
{
    png->offset = offset;
    int i; for(i=0; i < N; ++i)
        png->val[i] = vals[i];

    return png;
}

png_t* png_from_vector(int n, vector_t* vector)
{
    int i;
    png_t* ret = png_create();
    ret->offset = n;
    ret->match  = 0;
    for(i=0; i < N; i++)
        ret->val[i] = vector->val[i];

    return ret;
}

int char_index(char* str, char c)
{
    int i;
    for(i=0; i < strlen(str); i++) 
        if( str[i] == '\0' )
            return -1;
        else if( str[i] == c )
            return i;

    return -1;
}

png_t* read_png(FILE* file)
{
    char        line[256];
    png_t*      png;
    int         i;
    int         ignore;
    char*       inc;

    read_line(line, file); 
    png = png_create();
    
    sscanf(line, "%d %d", &ignore, &(png->offset));

    inc = line;
    inc += char_index(line, ' '); inc++;
    inc += char_index(inc, ' '); inc++;
    for(i=0; i < N; i++) {
        sscanf(inc, "%x", &(png->val[i]));
        inc += char_index(inc, ' '); inc++;
    }

    for(i=0; i < MAX_TIME; i++)
        png->match = simtime;

    return png;
}

int min(int x, int y)
{
    if( x > y )
        return y;

    return x;
}

float percent(int x, int y)
{
    return(100.0*x/y);
}


unsigned int hashfn(void* val)
{
    unsigned int hash = 0;
    png_t* png = (png_t*)val; 
    u_char* vals = (u_char*)malloc(sizeof(int) + sizeof(u_char)*N);
    int i;

    // Copy integer (4 bytes) to png val
    vals[0]= ((u_char*)val)[0];
    vals[1]= ((u_char*)val)[1];
    vals[2]= ((u_char*)val)[2];
    vals[3]= ((u_char*)val)[3];

    for(i=0; i < N; i++)
        vals[i+4] = png->val[i];

    for(i=0; i < N+4; i++)
        hash = vals[i] + (hash << 6) + (hash << 16) - hash;

    free(vals);
    return hash;
}

guint png_hash_fn(gconstpointer key) {
  return hashfn((void*)key);
}

unsigned int flow_fn(void* val)
{
    unsigned int hash = 0;
    u_char* vals = (u_char*)val;
    int i;

    for(i=0; i < sizeof(flow_t); i++)
    {
        if(!(i==2 || i==3 || i==4))
            hash = vals[i] + (hash << 6) + (hash << 16) - hash;
    }
    return hash;
}

boolean hasheqn(void* val1, void* val2)
{
    png_t* p1 = (png_t*)val1,
          *p2 = (png_t*)val2;
    int i;

    if( p1->offset != p2->offset)
        return false;

    for(i=0; i < N; i++)
        if( p1->val[i] != p2->val[i] )
            return false;

    return true;
}

gint png_hash_cmp(gconstpointer a, gconstpointer b) {
  return hasheqn((void*)a, (void*)b);
}


png_t* png_clone(png_t* png)
{
    int i;
    png_t* ret = png_create();
    ret->offset = png->offset; 

    for(i=0; i < N; i++)
        ret->val[i] = png->val[i];

    ret->match   = png->match;

    return ret;
}


node_t* node_head()
{
    return head;
}

node_t* node_default()
{
    return node_default_find();
}

node_t* node_find(node_t* node, int id)
{
    node_t* ret = NULL;
    if(node == NULL) return NULL;
    if(node->id == id)
        return node;

    if(node->type == QUEUE) return NULL;

    ret = node_find(node->left, id);
    if(ret == NULL) ret = node_find(node->right, id);

    return ret;
}

boolean throughput_in_range(double throughput)
{
    double midpoint = old_algorithm ? 0.3 : 0.5; // These should really be in constants.h
    double min      = midpoint - RANGE;
    double max      = midpoint + RANGE;

    if( (throughput > min) &&      // Minimum proportion of packets to split
        ( throughput < max) )
        return true;
    else
        return false;
}

void hash_png_free(void* png)
{
    png_free((png_t*) png);
}


static
void delete_png(gpointer key, gpointer value, gpointer data) {
    png_free((png_t*) key);        
}

void GLIB_destroy_pngs(GHashTable* pngs) {
    g_hash_table_foreach(pngs, delete_png, NULL);
}

/* This is a hack used to import some variables into
   split_sufficient while only using one pointer variable.
*/
struct split_env {
    int size;
    gboolean* found;
    png_t**   ret;
};

void split_sufficient(gpointer key, gpointer value, gpointer data) {
//    fprintf(stderr, "ENTERING split_sufficient\n");
    png_t* png      = (png_t*) key;
    int    count    = GPOINTER_TO_INT(value);
    struct split_env* user = (struct split_env*)data;

    if(*(user->found) == FALSE) {
        double throughput = 1.0*(count)/user->size;
        if(throughput_in_range(throughput)) {

            *(user->ret)   = png_clone(png);
            *(user->found) = TRUE;

            if(config->verbose) {
                fprintf(flog, "SPLITTING:  val: %d size %d throughput: %f\n", count, user->size, throughput);
                fprintf(flog, "PNG is ");
                png_print(flog, *(user->ret));
                fprintf(flog, " with ptr %d\n", *(user->ret));
            }
        }    
    }
}
 
png_t* GLIB_png_find_sufficient(GHashTable* pngs, int size) {
    png_t* ret = NULL;
    gboolean found = FALSE;
    struct split_env env = { size, &found, &ret };

    //fprintf(stderr, "ENTERING GLIB_png_find_sufficient\n");

    if(config->verbose)
        fprintf(flog, "PNG split with sufficient\n    PNG count:    %d\n    Packet Count: %d\n", g_hash_table_size(pngs), size);

    g_hash_table_foreach(pngs, split_sufficient, &env);

    // fprintf(stderr, "EXITING GLIB_png_find_sufficient\n");
    return ret; 
}

png_t* GLIB_png_find_optimal(GHashTable* pngs, int size) {
    png_t* ret = NULL;
    double difference = DBL_MAX;

    if(config->verbose)
        fprintf(flog, "PNG split with sufficient\n    PNG count:    %d\n    Packet Count: %d\n", g_hash_table_size(pngs), size);

    void split_optimal(gpointer key, gpointer value, gpointer user) {
        png_t* png      = (png_t*) key;
        int    count    = GPOINTER_TO_INT(value);

        double throughput = fabs(1.0*count/size - 0.5);
        if(throughput < difference) {
            difference = throughput;
            ret = key; 
        }
    } 

    int count = GPOINTER_TO_INT(g_hash_table_lookup(pngs, ret));
    int throughput = 1.0*count/size;
    if(throughput_in_range(throughput))
        return png_clone(ret);
    else
        return NULL;
} 
        
int sum_max_simtime(node_t* q)
{
    int total=0, i;
    assert(q->type == PNG_OP || q->type == QUEUE);

    for(i=0; i < MAX_TIME; i++)
        total += q->count_cache[i];

    assert(total >= 0);

    return total;
}

void age_pngs(node_t* node)
{
    int i;
    png_op_t* pngs;
    
    // Ensure that only png_ops are operated upon.
    assert(node != NULL);
    if(node->type == QUEUE) return;
    assert(node->type == PNG_OP);

    pngs = &(node->data.pngs);

    // if (pngs->png_size > 0)         // Don't eliminate pngs if they're the only thing there.
    for(i=0; i < pngs->png_size; i++)
    {
        int val = simtime - pngs->pngs[i]->match;
        //printf("%d\n", val);
        if( val > MAX_TIME)
        {
            fprintf(flog, "AGING OUT PNG in n%d.\n", node->id);
            if(i != pngs->png_size-1)
            {
                png_free(pngs->pngs[i]);
                pngs->pngs[i] = pngs->pngs[pngs->png_size-1];
                i = i-1;
            }
            pngs->pngs[pngs->png_size-1] = NULL;
            pngs->png_size--;
        } 
    }

    age_pngs(node->left);
    age_pngs(node->right);
}

double throughput(node_t* queue, int total)
{
    int i;
    int sum = 0;

    double ret;

    for(i=0; i < MAX_TIME; i++)
        sum += queue->count_cache[i];

    ret = 100.0 * sum / total;
    return ret;
}

queue_t* new_queue()
{
}

void sanity_check_png(png_t* png)
{
    if (png->offset < 0 || png->offset > 1500-N)
    {
        fprintf(stdout, "Illegal offset: %d\n", png->offset);
        exit(1);
    }
}

void node_init_caches(node_t* node)
{
    assert(node != NULL);

    node->count_cache[simtime % MAX_TIME] = 0;
    if(node->type == QUEUE)
    {
        //init_packet_cache(NODE2QUEUE(node), true);
        // FIX THIS BY ADDING VISITOR PATTERNS TO THINGS.
        p_stats_init(&(node->data.queue.stats[simtime % MAX_TIME]));
        return;
    }
    node_init_caches(node->left);
    node_init_caches(node->right);
}

void init_count_cache()
{
    packet_counts[simtime % MAX_TIME] = 0;
    node_init_caches(head);
}

int mature_packet_count()
{
    return max_simtime_packets;
}

void calc_packet_totals()
{
    int i;

    max_simtime_packets = 0;
    total_packets += tick_packet_count;

    for(i=0; i < MAX_TIME; i++) 
        max_simtime_packets += packet_counts[i];

    fprintf(stdout, "    Totals: %d %d %d %d %d\n", simtime, tick_packet_count, max_simtime_packets, total_packets, num_queues);
}

int packet_count(node_t* node)
{
    int i;
    int sum = 0;
    assert(node != NULL);
    assert(node->type == QUEUE);

    for(i=0; i < MAX_TIME; i++)
        sum += node->count_cache[i];
    printf("QUEUE PACKETS: %d\n", sum);
    return sum;
}

double rate(node_t* node)
{
    printf("TOTAL PACKETS: %d\n", max_simtime_packets);
    return 100.0 * (double)packet_count(node) / (double)max_simtime_packets;
}

boolean node_frozen(node_t* node)
{
    return(node->frozen);
}

void node_freeze(node_t* node)
{
    node->frozen = true;
}    

node_t* node_png_create_with_png_op_id(png_op_t* pnglist, node_t* parent, int id)
{
    int i;
    node_t* node = (node_t*)malloc(sizeof(node_t));

    node->frozen = false;
    node->born = simtime;
    node->type = PNG_OP;

    node->data.pngs.png_size = pnglist->png_size;
    for(i=0; i < MAX_PNGS; i++)
        node->data.pngs.pngs[i] = pnglist->pngs[i];

    for(i=0; i < MAX_TIME; i++)
        node->count_cache[i] = 0;

    node->parent = parent;
    node->id = id;
    return node;
}

node_t* node_png_create_with_id(png_t* png, node_t* parent, int id)
{
    int i;
    node_t* node = (node_t*)malloc(sizeof(node_t));
    node->frozen = false;
    node->born = simtime;
    node->type = PNG_OP;
    node->data.pngs.png_size = 1;

    for(i=0; i < MAX_PNGS; i++)
        node->data.pngs.pngs[i] = NULL;
    node->data.pngs.pngs[0] = png;

    for(i=0; i < MAX_TIME; i++)
        node->count_cache[i] = 0;

    node->parent = parent;
    node->id = id;
    return node;
}

node_t* node_png_create(png_t* png, node_t* parent)
{
    int id = id_get_next();
    return node_png_create_with_id(png, parent, id);
}

boolean queue_splittable(node_t* node)
{
    double rate;
    assert(node != NULL && node->type == QUEUE);
    if(node_frozen(node))
    { 
        printf("NODE n%d IS FROZEN!\n", node->id);
        return false;
    }

    rate = throughput(node, max_simtime_packets);
    if(rate > 2.0 && MATURE(node))
        return true;

    return false;
}


void node_update_child(node_t* parent, node_t* old_child, node_t* new_child)
{
    if(parent == NULL)      // Magic to make head work properly
    {
        head_set(new_child);
    }
    else
    {
        printf("Parent: n%d old: n%d new: n%d\n", parent->id, old_child->id, new_child->id);
        assert(parent->type == PNG_OP);
        assert(parent->left == old_child || parent->right == old_child);

        if(parent->left  == old_child) parent->left  = new_child;
        if(parent->right == old_child) parent->right = new_child;
    }
}


boolean split_queue_old(node_t* node)
{
    png_t*  png;        // The new png
    node_t* ret;        // The new node
    node_t* parent;
   
    printf("Attempting to split node n%d\n", node->id);
    printf("Head is n%d", head->id);
    
    // Asserts
    assert(node != NULL);
    assert(node->type == QUEUE);
    assert(node->parent != NULL || node == node_head()); 
    
    // The old infrastructure
    // png = png_find(NODE2QUEUE(node));

    png = packet_cache_png_find(cache, NODE2QUEUE(node));

    // Log reports
    if (png == NULL) {
        if(config->verbose)
            fprintf(flog, "UNABLE TO SPLIT QUEUE n%d.\n", node->id);
        return false;
    }

    if(config->verbose) {
        fprintf(flog, "SPLITTING QUEUE N%d at simtime %d with png: ",
                node->id, simtime);
        png_print(flog, png);
        fprintf(flog, "\n");
    }

    if(node->parent == NULL)
    {
        // Initialize the new node and its leafs
        ret = node_png_create(png, NULL);
        ret->left = node;

        ret->right = node_queue_create(ret);
        node->parent = ret;
        node_queue_init(node, true);

        // Fix up head
        head_set(ret);
    }
    else if(node == node_default())
    {
        ret = node_png_create(png, node->parent);
        ret->left = node_queue_create(ret);
        ret->right = node;
        node_queue_init(node, true);

        node_update_child(node->parent, node, ret);
        node->parent = ret;
    }
    else if(node->parent->parent == NULL)
    {
        ret = node_png_create(png, NULL);
        ret->left = node_queue_create(ret);
        ret->right = node->parent;
        node->parent->parent = ret;
        node_queue_init(node, true);

        // Fix up head
        head_set(ret);
    }
    else 
    {
        // For abbreviation...
        parent  = node->parent->parent;

        // Fix up ret's pointers
        ret     = node_png_create(png, parent);
        ret->left = node_queue_create(ret);
        ret->right = node->parent;

        // Put ret in the tree
        parent->right = ret;

        // Fix the leaf's parent.
        ret->right->parent = ret;

        // Reinit the queue
        node_queue_init(node, true);

        // Reinit the queue's png as well
        ret->parent->born = simtime;
    }

    return true;
}

boolean split_queue_new(node_t* node)
{
    png_t* png;
    node_t* old;
    assert(node != NULL);
    assert(node->type == QUEUE);

    // Old infrastructure
    //png = png_find(&(node->data.queue));

    png = packet_cache_png_find(cache, NODE2QUEUE(node));

    assert(node->parent != NULL || node == head);

    if (png == NULL) {
        if(config->verbose) fprintf(stdout, "UNABLE TO SPLIT QUEUE!\n");
        return false;
    }


    if(node != node_head()) {
        fprintf(flog, "Parent is: n%d\nSplitting with ", node->parent->id);
        png_print(flog, png);
        fprintf(flog,"\nParent is ");
        node_print(flog, node->parent);
        fprintf(flog, "\n");
        if(node_png_contains(node->parent, png)) {
            packet_cache_show_pngs_for_offset(cache, NODE2QUEUE(node), png->offset);
            if (fgraph) {
                print_graph(fgraph);
                fclose(fgraph);
            }
            exit(1);
        }
    }

    if(config->verbose) {
        printf("SPLITTING QUEUE N%d at simtime %d with png: ",
               node->id, simtime);
        png_print(stdout, png);
        printf("\n");
    }
 
    {
	node_t* new_node = node_png_create(png, node);

        if(config->verbose)
            fprintf(flog, "The new png node is %x, id is n%d.\n",
                    new_node, new_node->id);
        if(node == node_head()) {
            if(config->verbose)
                printf("Setting node n%d to head.\n", new_node->id);
            head_set(new_node); 
        }
        else if(node->parent->left == node) {
            if(config->verbose)
                printf("Setting node: n%d->left to n%d\n",
                       node->parent->id, node->id);
            node->parent->left = new_node;
        } 
        else if(node->parent->right == node) {
            if(config->verbose)
                printf("Setting node: n%d->right to n%d\n",
                       node->parent->id, node->id);
            node->parent->right = new_node;
        }
        else {
            printf("ERROR!"); exit(EXIT_FAILURE);
        }

        new_node->parent = node->parent;
        new_node->left = node_queue_create(new_node);
        new_node->right = node_queue_create(new_node);

        node_queue_free(node);
        node = NULL;

        if(new_node != head)
            assert(new_node->parent != NULL && new_node->parent != node);

        assert(new_node->left != NULL && new_node->right != NULL);
        assert(new_node->left != node && new_node->right != node);

        return true;
    }
/*
    old = node;

    node = node_png_create(png, node); 
    node->left = NULL;
    node->right = NULL;
    node_update_child(old->parent, old, node);

    node->left = node_queue_create(node);

    if(old_algorithm) node_freeze(node->left);

    node->right = old;
    node->parent = old->parent;

    node_queue_init(node->right, true);
    node->right->parent=node;

    assert(node != node->right || node != node->left &&
           node != node->parent);

    return true;
*/
 
}

boolean split_queue(node_t* node)
{
    if(old_algorithm)
    {
        split_queue_old(node);
    }
    else
    {
        split_queue_new(node);
    }

}

/* Shouldn't we actually we looking at queue_splittable, etc?) */
void split_queues(node_t* node)
{
    assert(node != NULL);
    assert(node->type == PNG_OP || node->type == QUEUE);

    if(node->type == PNG_OP)
    {
        split_queues(node->left);
        split_queues(node->right);
    }
    else // this must be QUEUE type
    {
        double rate = throughput(node, max_simtime_packets);

        if(!node_frozen(node) && rate > 2.0 && MATURE(node))
        {
            boolean split = split_queue(node);
        }
    }
}


node_t* find_default(node_t* node)
{
    assert(node != NULL);
    assert(node->type == PNG_OP || node->type == QUEUE);
    if(node->type == QUEUE)
        return node;
    else return find_default(node->right);
}

node_t* node_default_find()
{
    return find_default(head);
}

boolean queue_mergable(node_t* node)
{
    double rate;
    assert(node != NULL);
    assert(node->type == QUEUE);
    rate = throughput(node, max_simtime_packets);

    if((rate <= MAX_MERGE_RATE) && MATURE(node))
    {
        printf("We should merge.\n");
        return true;
    }

    return false;
}

int node_png_match_simtime(node_t* node)
{
    int i, max=0;
    assert(node != NULL);
    assert(node->type == PNG_OP); 

    
    for(i=0; i < node->data.pngs.png_size; i++)
        if(node->data.pngs.pngs[i]->match > max)
            max = node->data.pngs.pngs[i]->match; 

    return max;
}

boolean png_mergable(node_t* node)
{
    double rate;
    assert(node != NULL);
    assert(node->type == PNG_OP);
    rate = throughput(node, max_simtime_packets);

    if( rate <= MAX_MERGE_RATE )
    {
        printf("PNG IS MERGABLE!\n");
        printf("%d n%d %d %d\n", simtime, node->id, node_png_match_simtime(node), (simtime - node_png_match_simtime(node) > MAX_TIME));
    }

    if((rate <= MAX_MERGE_RATE) && MATURE(node))
    {
        printf("PNG n%d IS MERGABLE!\n", node->id);
        return(true);
    }

    return(false);
}

/* How do we want to handle node merging?
 * by match dates or by counts?
 */
boolean node_mergable(node_t* node)
{
    double rate;
    boolean ret;

    assert(node != NULL);
    assert(node->type == PNG_OP || node->type == QUEUE);

    if(!MATURE(node)) return false;

    if(node->type == QUEUE)
        ret = queue_mergable(node);
    else
        ret = png_mergable(node);

    if(ret && config->verbose)
        fprintf(stdout, "NODE N%d (%d) is mergable\n", node->id, node);

    return(ret);
}

void node_queue_free(node_t* node)
{
    int i;
    assert(node != NULL && node->type == QUEUE);

    //init_packet_cache(NODE2QUEUE(node), true);          // Free Packets

    for(i=0; i < num_queues; ++i)
    { 
        if(queues[i] == node)
        {
            queues[i] = queues[num_queues-1];
            queues[num_queues-1] = NULL;
            break;
        }
    }

    packet_cache_free(cache, NODE2QUEUE(node));

    //free(node);  // We should do this!
    num_queues--;
}


void png_free(png_t* png)
{
    free(png->val);
    free(png);
}


void node_png_free(node_t* node)
{
    int i;
    assert(node != NULL && node->type == PNG_OP);
/*
    assert(node->data.pngs.png_size > 0 && 
           node->data.pngs.png_size < MAX_PNGS);
*/

    for(i=0; i < node->data.pngs.png_size; i++)
        png_free(node->data.pngs.pngs[i]);

    free(node);
}

void node_free(node_t* node)
{
    assert(node != NULL);

    if(node->type == QUEUE)
    {
        printf("Removing Queue N%d - %d %d\n", node->id, num_queues, num_queues);
        //queues_dec(node);
        //node_queue_free(node);
        node_queue_free(node);
        return;
    }

    assert(node->type == PNG_OP && node->left != NULL && node->right != NULL);
    node_free(node->left);
    node_free(node->right);

    node_png_free(node);
}

#define LEFT 0
#define RIGHT 1

void node_clip_tree(node_t* node, int direction)
{
    assert(direction == 0 | direction == 1);

    fprintf(flog, "DELETING subtree n%d ->left\n", node->id);

    if(old_algorithm && node != node_head()) png_op_copy( &(node->parent->data.pngs), &(node->data.pngs));

    node_update_child(node->parent, node, direction ? node->left : node->right );

    if(direction == LEFT)
    {
        node->right->parent = node->parent;
        node_free(node->left); // Shouldn't this free node, not node->left?
    }
    else 
    {
        node->left->parent = node->parent;
        node_free(node->right);
    }
}

void node_merge(node_t* node)
{
    assert(node != NULL);

    if(node->type == QUEUE) return;

    assert(node->left != NULL && node->right != NULL && node->type == PNG_OP);

    if(node_mergable(node->left))
        node_clip_tree(node, LEFT);
    else
        node_merge(node->left);  

    if(node_mergable(node->right))    
        node_clip_tree(node, RIGHT);
    else
        node_merge(node->right);
}

#undef LEFT
#undef RIGHT

void tick()
{
    time_t cur_time; 
    time(&cur_time);
 
    fprintf(flog, "TICK() - %d\n\n", cur_time);
    // AGEING ---------------------------------------------------------
    if(ALLOW_AGEING) age_pngs(head);

    // MERGE QUEUES --------------------------------------------------
    if(ALLOW_MERGING) node_merge(head);

    // SPLIT QUEUES --------------------------------------------------
    if(ALLOW_SPLITTING) split_queues(head);  // Direction shouldn't matter here.

    tick_packet_count = 0;
}


int png_match(png_t* png, const u_char* packet, unsigned int len)
{
    int i;

#ifdef VARIABLE_N
    fprintf(stderr, "NOT IMPLEMENTED YET!\n");
    exit(0);
#endif

/*
    if((png->offset + N) >= packet->ip_len)
        return false;

    for(i=0; i < N; ++i)
    {
        if( packet->ip_packet[i+png->offset] != png->val[i] )
            return false;
    }

    png->match = simtime;
*/

    if((png->offset + N) >= len) // This appears to be a bug.
    if((png->offset + N) > len)
        return false;

    for(i=0; i < N; ++i)
        if(packet[i+png->offset] != png->val[i])
            return false;

    return true;
}

/* This is pretty ugly.  Why have queues_print and print_queue_array?
 */
void print_end_of_scan_packets(char* begin, char* end)
{
    // MOVE to where this is called, instead of here.
    if(!config->by_file) fprintf(flog, "Packets from %s to %s\n", begin, end);

    if(fgraph) { print_graph(fgraph); fflush(fgraph); }

    if(globals) print_stats(flog);
    queues_print(flog);
    p_stats_print(flog, &global_stats);
    if(config->verbose) print_queue_array(flog);
    fflush(flog);
}

int jim = 0;

void scan_packet(u_char *user, const struct pcap_pkthdr *h,
                 const u_char *packet)
{
    int match_id, i;

    assert(h->caplen == h->len);

    classify(&global_stats, packet, h->caplen);
    classify(&(tick_stats[simtime % MAX_TIME]), packet, h->caplen);

/*
    if(packet->fragments != NULL)
        printf("PACKET HAS FRAGMENTS!\n");
*/
     
    // Move this into a separate function?
    // Probably Yes!
    if(!config->by_file)
    {
        char* tmp;
        char* ts; 
        char* lt;

        if(last_tick == 0) {
            last_tick = h->ts.tv_sec;
            if (db_tree) db_create_new_tick(db_tree, h->ts.tv_sec,
                                            queue_counter-1, png_counter-1);
        }

        tmp = ctime((time_t*)&(last_tick));
        ts = copystr(tmp);
        tmp = ctime(&(h->ts.tv_sec));
        lt = copystr(tmp);

        ts[strlen(ts)-1] = '\0';
        lt[strlen(lt)-1] = '\0';

        // ( x min * 60 sec/min )
        if(h->ts.tv_sec - last_tick >= TICK_INTERVAL*60) 
        {
            calc_packet_totals();
            if (db_tree) db_create_new_tick(db_tree, h->ts.tv_sec,
                                            queue_counter-1, png_counter-1);
            print_end_of_scan_packets(ts, lt);
            tick();
            sanity();
            simtime++;
            last_tick = h->ts.tv_sec;
            p_stats_init(&(tick_stats[simtime % MAX_TIME]));
            init_count_cache();
        }
        free(ts); free(lt);
    }

    /* check what we should be capturing every 2 seconds */
    if (db_tree)
        if (h->ts.tv_sec - last_capture_check > 2 || !last_capture_check) {
            db_tree_lookup_capture_requests(db_tree, pcap);
            last_capture_check = h->ts.tv_sec;
        }

    match_id = match(head, packet, h->caplen);

    if (db_tree)
        for (i = 0; i < db_tree_num_capture_requests(db_tree); ++i) {
            db_capture_t *cap = db_tree_capture_request(db_tree, i);
            if (db_capture_node_id(cap) == match_id) {
                /* capture this packet to the database */
                db_capture_packet(cap, h, packet);
            }
        }

    tick_packet_count++;
    packet_counts[simtime % MAX_TIME]++;
}

int find_ip_offset(pcap_t *pcap)
{
    int type = pcap_datalink(pcap);
    int o = 0;

    switch(type) {
        /* these offsets are copied from the qcap library.. */
    case DLT_EN10MB: /* Ethernet - 6o src addr, 6o dst addr, 2o ?, 2o ethtype*/
        o = 14;
        break;
    case DLT_CHDLC: /* Cisco HDLC - 1o address, 1o spare, 2o ethtype */
        o = 4;
        break;
    case DLT_RAW: /* raw IP - no DL */
        o = 0;
        break;
    case DLT_ATM_RFC1483: /* 3o llc, 3o oui, 2o ethtype */
        o = 8;
        break;
    default:
        fprintf(stdout, "Unknown link type %d\n", type);
        exit(1);
        break;
    }
    return o;
}

void process_file(const char* filename)
{
    int         ret;
    char        errbuf[PCAP_ERRBUF_SIZE];

    fprintf(stdout, "Processing: %s\n", filename);
    pcap = pcap_open_offline(filename, errbuf); 
    if(pcap == NULL) {
        fprintf(stdout, "ERROR: %s\n", errbuf);
        exit(1);
    }
    ip_offset = find_ip_offset(pcap);

    if (!end_loop)
        if (pcap_loop(pcap, -1, scan_packet, NULL) == -1) {
            fprintf(stdout, "ERROR: %s\n", pcap_geterr(pcap));
            exit(1);
        }

    pcap_close(pcap);
    pcap = NULL;
    fprintf(stdout, "Finished processing: %s\n", filename);
}


void print_pngs(FILE* file, png_t* png)
{
    int i;
    fprintf(file, "%d,", png->offset);
    for(i=0; i < N; i++)
    {
        fprintf(file, " 0x%.2x", png->val[i]);
    }
}

void png_print(FILE* file, png_t* png)
{
    assert(file != NULL);
    assert(png != NULL);
    print_pngs(file, png);
}

void pngs_print(FILE* file, png_op_t* pngs)
{
    int i;
    if (pngs->png_size > 0)
    {
        for(i=0; i < pngs->png_size-1; i++)
        {
            png_print(file, pngs->pngs[i]);
            fprintf(file, " || ");
        }
        png_print(file, pngs->pngs[pngs->png_size-1]);
    }
}


int pngs_to_head(node_t* node)
{
    int depth=0;
    assert(node != NULL && node->type == QUEUE);

    node = node->parent;
    while(node != NULL)
    {
        depth++;
        node = node->parent;
    }

    return depth;
}

int max_depth()
{
    int max = 0;
    int i;
    for(i=0; i < num_queues; i++)
    {
        int tmp = pngs_to_head(queues[i]);
        if(tmp > max) max = tmp;
    }

    return max;
}

int packets()
{
    int i;
    int ret = 0;
    for(i=0; i < num_queues; i++)
        ret += LAST_TICK(queues[i]);

    return ret;
}

int comparisons()
{
    int i;
    int sum = 0;
    for(i=0; i < num_queues; i++)
    {
        int tmp = pngs_to_head(queues[i]);
        sum += tmp*LAST_TICK(queues[i]);
    }
    return sum;
}

int total_depth(node_t* node, int cur)
{
    int left, right;
    assert(node != NULL);

    if(node->type == QUEUE) return(cur+1);

    left = total_depth(node->left, cur+1);
    right = total_depth(node->right, cur+1);

    return left+right;
}

int total_packet_depth(node_t* node, int cur)
{
    int left, right;
    assert(node != NULL);

    if(node->type == QUEUE) return (node->count_cache[simtime % MAX_TIME]*cur);

    left  = total_depth(node->left, cur+1);
    right = total_depth(node->right, cur+1);

    return(left+right);
}

float avg_depth()
{
    int sum = total_depth(head, 0);
    return((1.0*sum)/num_queues);
}

float avg_packet_depth()
{
    int sum = total_packet_depth(head, 0);

    return((1.0*sum)/tick_packet_count);
}

int node_queues_mergable_count()
{
    int i;
    int sum = 0;
    for(i=0; i < num_queues; i++)
        if(queue_mergable(queues[i]))
        { //  fprintf(stderr, "Time %d N%d\n", simtime, queues[i]->id);
            sum++; }

    return sum;
}

int node_queues_splittable_count()
{
    int i;
    int sum = 0;
    for(i=0; i < num_queues; i++)
        if(queue_splittable(queues[i]))
            sum++;

    return sum;
}

int node_queues_immature_count()
{
    int i;
    int sum = 0;
    for(i=0; i < num_queues; i++)
        if(!MATURE(queues[i]))
            sum++;

    return sum;
}

// Is the node we've found really there?
node_t* find_node(node_t* node, node_t* seek)
{
    node_t* ret = NULL;
    assert(node != NULL);
    if(node == seek) return node;
 
    if(node->type == QUEUE) return NULL;

    ret = find_node(node->left, seek);
    if(ret == NULL) ret = find_node(node->right, seek);
    return ret;
}

void scan_sanity(node_t* node)
{
    assert(node != NULL);
    if(node->type == QUEUE)
    {
        queues[num_queues] = node;
        num_queues++;
    }
    else
    {
        scan_sanity(node->left);
        scan_sanity(node->right);
    }
}

void regain_sanity()
{
    num_queues = 0;
    scan_sanity(node_head());
}

void sanity()
{
    int i;
    boolean fail = false;
    for(i=0; i < num_queues; i++)
    {
        node_t* node = find_node(head, queues[i]);
        if(node == NULL)
        {
            fprintf(stderr, "QUEUES INCONSISTENT: Time: %d N%d\n", simtime, queues[i]->id);
            fprintf(stderr, "N%d->parent = n%d", queues[i]->id, queues[i]->parent->id);
            exit(0);
            fail = true;
        }
    }
    if(fail) exit(0);
}

void print_stats_summary(FILE* file)
{
    p_stats_print_summary(file, &global_stats);
}

void print_stats(FILE* file)
{
    int splittable = node_queues_splittable_count();
    int mergable   = node_queues_mergable_count();
    int immature   = node_queues_immature_count();

    int mono_nodes = 0;
    int mono_packets = 0;

    visit_queues(node_head(), count_mono_nodes, &mono_nodes);
    visit_queues(node_head(), total_mono_packets, &mono_packets);

    printf("Time: %d -------------------------\n", simtime);
    printf("GLOBAL queues:          %d\n", num_queues);
    printf("GLOBAL normal:          %d\n", num_queues - (splittable + mergable + immature));
    printf("GLOBAL immature:        %d\n", immature);
    printf("GLOBAL splittable:      %d\n", splittable);
    printf("GLOBAL mergable:        %d\n", mergable);
    printf("GLOBAL max depth        %d\n", max_depth());
    printf("GLOBAL packets:         %d\n", tick_packet_count);
    printf("GLOBAL comparisons      %d\n", comparisons());
    printf("GLOBAL comp/pkt         %f\n", 1.0*comparisons()/packets());
    printf("GLOBAL mono queues      %d\n", mono_nodes);
    printf("GLOBAL mono packets     %d\n", mono_packets);
}

void print_params(FILE* file)
{
    fprintf(file, "Parameters --------------------------------------------------------------------\n");
    fprintf(file, "VERSION        %s\n",   VERSION);
    fprintf(file, "AGEING         %s\n",   ALLOW_AGEING    ? "on" : "off");
    fprintf(file, "SPLITTING      %s\n",   ALLOW_SPLITTING ? "on" : "off");
    fprintf(file, "MERGING        %s\n",   ALLOW_MERGING   ? "on" : "off");
    fprintf(file, "IGNORE PORT    %s\n",   config->ignore_port ? "on" : "off");
    fprintf(file, "HEADER OFFSET  %d\n\n", config->noheaders);
    fprintf(file, "Constants ---------------------------------------------------------------------\n");
    fprintf(file, "TICK_INTERVAL  %d (minutes in each data file)\n", TICK_INTERVAL);
    fprintf(file, "MAX_TIME       %d (Number of TICK_INTERVALS used for merging/splitting decisions\n", MAX_TIME);
    fprintf(file, "MAX_PNGS       %d (maximum number of pngs per decision node)\n", MAX_PNGS);
//    fprintf(file, "MAX_PACKETS    %d (maximum number of packets in the queue packet cache)\n", MAX_PACKETS);
    fprintf(file, "MAX_MERGE_RATE %.2f (rate at or below which queues are deleted)\n", MAX_MERGE_RATE);
    fprintf(file, "\n");
    fprintf(file, "Time:          %d\n", simtime);
}

void print_queue_array(FILE* file)
{
    int i;
    for(i=0; i < num_queues; i++)
    {
        node_t* node = queues[i];
        assert(node != NULL && node->type == QUEUE);

        fprintf(file, "QUEUE N%d: %d\n", node->id, node);
    }
}







void node_png_print(FILE* file, node_t* node)
{
    int i;
    fprintf(file, "PNG n%d %d: ", node->id, node->born);
    for(i=0; i < node->data.pngs.png_size; i++)
        png_print(file, node->data.pngs.pngs[i]);
    fprintf(file, " %d %d\n", node->count_cache[simtime % MAX_TIME],
                              sum_max_simtime(node));
}

void node_queue_print_packets(FILE* file, node_t* node)
{
    int i;
    assert(node != NULL);

    printf("Not implemented for new packet cache.\n");
/*
    for(i=0; i < CACHE_SIZE(node); i++)
    {
        flow_t* flow = flow_from_packet(node->data.queue.packet_cache.packets[i]);
        if(flow != NULL)
        {
            flow_print(file, flow);
            free(flow);
        }
        else
            fprintf(file, " <NOT TCP | UDP>");
        fprintf(file, "\n");
    }
*/
}

void node_queue_print(FILE* file, node_t* node)
{
    int i;
    float max;
    float info;

    printf("Not implemented for new packet cache.\n");
/*
    
    info = entropy(CACHE_SIZE(node), node->data.queue.packet_cache.packets);
    max = logf(CACHE_SIZE(node))/logf(2);
    fprintf(file, "QUEUE n%d: <STACK HERE SOMEHOW> %d %d %d",
                  node->id,
                  node->born,
                  node->count_cache[simtime % MAX_TIME],
                  sum_max_simtime(node));
    fprintf(file, " %.2f %.2f\n", info, max);
    fprintf(file, "Packets:\n");

    node_queue_print_packets(file, node);
    node_queue_print_flows(file, node);
*/
}

void node_queue_print_flows(FILE* file, node_t* node)
{
    printf("Not implemented for the new packet cache.\n");
    assert(FALSE);
/*
    int i, flows;
    hashtable_t* dict;
    hashtable_itr_t* itr;

    dict = create_hashtable(4001, flow_fn, flow_eq);
    for(i=0; i < CACHE_SIZE(node); i++)
        dict_flow_packet_add(dict, node->data.queue.packet_cache.packets[i]); 

    flows = dict_flow_print(file, dict);
    fprintf(file, "N%d: %d flows.\n", node->id, flows);
*/
}

void node_print(FILE* file, node_t* node)
{
    assert(node != NULL && file != NULL);
    assert(node->type == PNG_OP || node->type == QUEUE);
    if(node->type == PNG_OP) node_png_print(file, node);
    else node_queue_print(file, node);
}

void process_end()
{
    if (pcap) pcap_breakloop(pcap);
    end_loop = true;
}

void process_live()
{
    int ret;
    char errbuf[PCAP_ERRBUF_SIZE];

    init_count_cache();

    fprintf(stdout, "Watching interface %s...\n", config->interface);

    pcap = pcap_open_live(config->interface, 65535, true, 0, errbuf);;
    if(pcap == NULL) {
        fprintf(stdout, "ERROR: %s\n", errbuf);
        exit(1);
    }
    ip_offset = find_ip_offset(pcap);

    if (!end_loop)
        if (pcap_loop(pcap, -1, scan_packet, NULL) == -1) {
            fprintf(stdout, "ERROR: %s\n", pcap_geterr(pcap));
            exit(1);
        }
   
    pcap_close(pcap);
    pcap = NULL;
}

void iterate()
{
    assert(datafile < config->numfiles);

    process_file(config->files[datafile]);

    if(config->by_file)
    {
        calc_packet_totals();
        print_end_of_scan_packets(NULL, NULL);
        tick();
        sanity();
        simtime++;
        datafile++;
        init_count_cache();
    }
    else
        datafile++;
}



// Where should this be used?
// IP_LEN - FIX
void packet_show(FILE* file, const u_char* packet, unsigned int len,
                 int start, int length)
{
    int i ;
    length = min(start+length, len-ip_offset);
    for(i=start; i < length-1; i++)
        fprintf(file, "0x%x ", packet[ip_offset+i]);
    if(length > start)
        fprintf(file, "0x%x", packet[ip_offset+i]);
} 

static vector_copy_contents(int* to, int* from, int size)
{
    int i;
    for(i=0; i < size; i++)
        to[i] = from[i];
}

void vector_add(vector_t* v, int n)
{
    if(v->max_size == v->size)
    {
        int* replacement = (int*)malloc(sizeof(int)*v->max_size*2);
        vector_copy_contents(replacement, v->val, v->max_size);
        free(v->val);
        v->val = replacement;
    }

    v->val[v->size++] = n;
}

#define VECTOR_INIT_SIZE 4
vector_t* vector_new(int n)
{
    vector_t* ret = (vector_t*)malloc(sizeof(vector_t));

    ret->max_size = VECTOR_INIT_SIZE;
    ret->val = (int*)malloc(sizeof(int)*VECTOR_INIT_SIZE);

    ret->val[0] = n;
    ret->size   = 1;

    return ret;
}
#undef VECTOR_INIT_SIZE

char buffer[1024];
int  count=1;

void printlog(FILE* file, const char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    char current[1024];

    vsnprintf(current, 1024, fmt, args);
    if(strncmp(buffer, current, 1024) == 0)
        count++;
    else
    {
        fprintf(file, "%d: %s", count, buffer);
        count=1;
        strncpy(buffer, current, 1024);
    }
}


png_op_t* png_op_create()
{
    int i;
    png_op_t* ret = (png_op_t*)malloc(sizeof(png_op_t));
    ret->png_size = 0;
    for(i=0; i < MAX_PNGS; i++)
        ret->pngs[i] = NULL;

    return ret;
}


void png_op_add(png_op_t* pngs, png_t* png)
{
    assert(pngs->png_size < MAX_PNGS);

    pngs->pngs[pngs->png_size++] = png;
}

void png_op_copy(png_op_t* dst, png_op_t* src)
{
    int i;
    for(i=0; i < src->png_size; i++)
    {
        png_t* png = png_clone(src->pngs[i]);
        png_op_add(dst, src->pngs[i]);
    }
}

void node_export(FILE* file, node_t* node)
{
    char* direction;
    assert(node->parent->left == node || node->parent->right == node);

    if(node->type == QUEUE)
        return;

    if(node->parent->left == node)
        direction = "left";
    else
        direction = "right";

    fprintf(file, "replace n%d.%s with n%d (", node->parent->id, direction, node->id);
    png_print(file, node->data.pngs.pngs[0]);
    fprintf(file, ")\n");
    node_export(file, node->left);
    node_export(file, node->right);
}

void tree_export(FILE* file)
{
    node_t* node = node_head();
    if(node->type != QUEUE)
    {
        fprintf(file, "replace n0 with n%d (", node->id);
        if(node->data.pngs.png_size == 0)
        {
            fprintf(stderr, "Node n%d has zero size.\n", node->id);
            exit(1);
        }
        else
        {
            png_print(file, node->data.pngs.pngs[0]);
            fprintf(file, ")\n");
        }
        node_export(file, node->left);
        node_export(file, node->right);
    }
}

int id_get_next()
{
    return png_counter++;
}

int id_queue_get_next() {
    return queue_counter++;
}

char* copystr(char* string)
{
//    char* ret = (char*)malloc(sizeof(char*)*4000);
    char* ret = (char*)malloc(sizeof(char)*(strlen(string)+1));
    strcpy(ret, string);
    return ret;
}

/*
int hashtable_sum(hashtable_t* table)
{
    struct hashtable_itr* itr;
    itr = hashtable_iterator(table);
    int sum = 0;

    do
    {
        int* val  = hashtable_iterator_value(itr);
        sum += *val; 
    }
    while(hashtable_iterator_advance(itr));

    free(itr);
    return sum;
}
*/

GHashTable* GLIB_packet2pngs(GHashTable* pngs, const u_char* packet,
                             unsigned int len)
{
//    fprintf(stderr, "ENTERING GLIB_packet2pngs\n");

    int i;

    if(pngs == NULL) {
        fprintf(stderr, "Creating a new Hashtable!\n");
        pngs = g_hash_table_new(png_hash_fn, png_hash_cmp);
    }

    if(config->ip_only && !isIPv4(packet, len)) return;
    
    for(i=config->noheaders; i < len - N; ++i) {
        int     j;
        int     val;
        char    png[N];
        png_t   tmp;
        png_t*  key = &tmp; 

        tmp.val = (unsigned char*)&png; // Kind of ugly.  Shouldn't we have an separate init function?
        tmp.match = simtime;
        tmp.offset = i;

        for(j=0; j < N; ++j)
            tmp.val[j] = packet[i+j];

        val = GPOINTER_TO_INT(g_hash_table_lookup(pngs, &tmp));
        if(!val) key = png_clone(&tmp);
        val++;

        g_hash_table_insert(pngs, key, GINT_TO_POINTER(val));
    }            

//    fprintf(stderr, "EXITING GLIB_packet2pngs\n");
    return pngs;
}


/*
hashtable_t* packet2pngs(hashtable_t* pngs, qcap_packet_t* packet)
{
    assert(packet->full_len == packet->full_caplen);

    int     i;
    int     length      = ip_only ? (packet->ip_len + 14 - N) : (packet->full_len) - N;
    u_char* contents    = packet->full_packet;

    if(pngs==NULL)
    {
        fprintf(stderr, "Creating a new Hashtable!\n");
        pngs = t_create_hashtable(16385, hashfn, hasheqn, hash_png_free, NULL);
    }

    if(ip_only && !isIPv4(packet)) return;

    for(i=noheaders; i < length; ++i)
    {
       int      j;
       int*     val;
       char     png[N];
       png_t    tmp;

       tmp.val = (unsigned char*)&png;
       tmp.match = simtime;
       tmp.offset = i;

       for(j=0; j < N; ++j)
           tmp.val[j] = contents[i+j]; 
      
       if((val = hashtable_search(pngs, &tmp)) == NULL)
       {
           val = (int*)malloc(sizeof(int));
           png_t* png = png_clone(&tmp);
           (*val) = 1;

           if(!hashtable_insert(pngs, png, val))
           {
               fprintf(stderr, "Hashtable insert failed!\n");
               exit(1);
           }
       }
       else
       {
           (*val)++;
       }
    }

    return pngs;
}
*/

boolean png_equals(png_t* one, png_t* two)
{
    int i;
  
// Fix for variable N
#ifdef VARIABLE_N
    assert(false);
#endif

    if(one->offset != two->offset)
        return false;


    for(i=0; i < N; i++)
        if(one->val[i] != two->val[i]) return false;

    fprintf(stderr, "HERE png_equals!\n");
    return true;
}

boolean node_png_contains(node_t* node, png_t* png)
{
    int i;
    png_op_t* pngs;
    assert(node != NULL && png != NULL);
    assert(node->type == PNG_OP);

    pngs = &(node->data.pngs);
    for(i=0; i < pngs->png_size; i++)
    {
        if(png_equals(pngs->pngs[i], png))
            return true;
    }

//    fprintf(stderr, "HERE node_png_contains!\n");
    return false;
}

void print_with_prefix(FILE* file, char* prefix, char* buffer)
{
    assert(file != NULL);
    assert(prefix != NULL);
    assert(buffer != NULL);

    while(buffer[0] != '\0')
    {
        fprintf(file, "%s", prefix);
        while(buffer[0] != '\n')
        {
            fputc(buffer[0], file);
            buffer++;
        }
        fputc('\n', file);
        buffer++;
    }
}

void visit_queues(node_t* node, void (*func)(node_t* node, void* data), void* data)
{
    assert(node->type == PNG_OP || node->type == QUEUE);
    if(node->type == PNG_OP)
    {
        visit_queues(node->left, func, data);
        visit_queues(node->right, func, data);
    }
    else
    {
        func(node, data);
    }
}
